@@ -15,7 +15,7 @@ use libseccomp::ScmpNotifResp;
 use nix::{errno::Errno, fcntl::AtFlags, unistd::AccessFlags, NixPath};
 
 use crate::{
-    fs::fdaccess,
+    fd::fdaccess,
     kernel::{syscall_path_handler, to_atflags},
     lookup::FsFlags,
     req::{PathArgs, SysArg, SysFlags, UNotifyEventRequest},
@@ -68,7 +68,7 @@ pub(crate) fn sys_faccessat2(request: UNotifyEventRequest) -> ScmpNotifResp {
     // SAFETY: Reject undefined/invalid/unused flags.
     let flags = match to_atflags(
         req.data.args[3],
-        AtFlags::AT_EACCESS | AtFlags::AT_EMPTY_PATH | AtFlags::AT_SYMLINK_NOFOLLOW,
+        crate::compat::AT_EACCESS | AtFlags::AT_EMPTY_PATH | AtFlags::AT_SYMLINK_NOFOLLOW,
     ) {
         Ok(flags) => flags,
         Err(errno) => return request.fail_syscall(errno),
@@ -120,7 +120,7 @@ fn syscall_access_handler(
 
     // We use MUST_PATH, dir refers to the file.
     assert!(
-        path.base.is_empty(),
+        path.base().is_empty(),
         "BUG: MUST_PATH returned a directory for access, report a bug!"
     );
     let fd = path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?;
