@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/kernel/chroot.rs: chroot(2) handler
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -32,24 +32,24 @@ pub(crate) fn sys_chroot(request: UNotifyEventRequest) -> ScmpNotifResp {
 
         if let Some(typ) = path.typ.as_ref() {
             if !typ.is_dir() {
-                return Ok(request.fail_syscall(Errno::ENOTDIR));
+                return Err(Errno::ENOTDIR);
             }
         } else {
-            return Ok(request.fail_syscall(Errno::ENOENT));
+            return Err(Errno::ENOENT);
         }
 
         // SAFETY: Do not allow nested chroots.
         if is_chroot {
-            return Ok(request.fail_syscall(Errno::EACCES));
+            return Err(Errno::EACCES);
         }
 
         // Acquire a write lock and chroot the sandbox.
         let mut sandbox = request.get_mut_sandbox();
         sandbox.chroot();
-        let verbose = sandbox.verbose;
+        let log_scmp = sandbox.log_scmp();
         drop(sandbox); // release the write-lock.
 
-        if verbose {
+        if log_scmp {
             warn!("ctx": "chroot_sandbox",
                 "sys": "chroot", "path": &path,
                 "msg": "change root approved",
