@@ -46,7 +46,7 @@ pub(crate) fn handle_getpeername(
             // and return EINVAL as necessary.
             let len = i32::from_ne_bytes(buf);
             let len = libc::socklen_t::try_from(len).or(Err(Errno::EINVAL))?;
-            if !is_valid_ptr(args[1]) {
+            if !is_valid_ptr(args[1], request.scmpreq.data.arch) {
                 // address length is positive however address is NULL,
                 // return EFAULT.
                 return Err(Errno::EFAULT);
