@@ -161,7 +161,7 @@ impl ThreadPool {
                 ctx.export_bpf(file)?;
 
                 // Aes worker rules
-                let ctx = AesWorker::prepare_confine(self.flags, &[], &[])?;
+                let ctx = AesWorker::prepare_confine(self.flags, &[], &[], true)?;
                 let file = OpenOptions::new()
                     .write(true)
                     .create_new(true)
@@ -205,7 +205,7 @@ impl ThreadPool {
                 rwrite!(stdout, "{}", seccomp_export_pfc(&ctx)?)?;
 
                 rwriteln!(stdout, "# Syd encryptor rules")?;
-                let ctx = AesWorker::prepare_confine(self.flags, &[], &[])?;
+                let ctx = AesWorker::prepare_confine(self.flags, &[], &[], true)?;
                 rwrite!(stdout, "{}", seccomp_export_pfc(&ctx)?)?;
             }
             _ => {}
