@@ -4,7 +4,7 @@
 //                If set is prctl, print the list of allowed prctl options
 //                If set is personality, print the list of allowed personalities.
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -70,6 +70,7 @@ syd::main! {
             println!("- futex");
             println!("- hook");
             println!("- keyring");
+            println!("- kill");
             println!("- mount");
             println!("- msgqueue");
             println!("- nice");
@@ -146,6 +147,11 @@ syd::main! {
                 println!("{name}");
             }
         }
+        Some("kill") => {
+            for name in syd::config::KILL_SYSCALLS {
+                println!("{name}");
+            }
+        }
         Some("mount") => {
             for name in syd::config::MOUNT_SYSCALLS {
                 println!("{name}");
