@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/utils/syd-pty.rs: Syd's PTY to STDIO bidirectional forwarder
 //
-// Copyright (c) 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -20,19 +20,21 @@ use nix::{
     sched::{unshare, CloneFlags},
     sys::{
         epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
-        prctl::set_dumpable,
+        resource::Resource,
         signal::{signal, sigprocmask, SigHandler, SigmaskHow, Signal},
         signalfd::{SfdFlags, SigSet, SignalFd},
-        termios::{cfmakeraw, tcgetattr, tcsetattr, OutputFlags, SetArg, Termios},
+        termios::{cfmakeraw, tcgetattr, tcsetattr, LocalFlags, OutputFlags, SetArg, Termios},
     },
-    unistd::{chdir, chroot, close, pipe2},
+    unistd::{chdir, chroot, pipe2},
 };
 use syd::{
-    compat::epoll_ctl_safe,
+    compat::{epoll_ctl_safe, set_dumpable},
     config::VDSO_SYSCALLS,
-    confine::{confine_mdwe, confine_scmp_madvise, extend_ioctl},
+    confine::{
+        confine_mdwe, confine_rlimit_zero, confine_scmp_madvise, extend_ioctl, CLONE_NEWTIME,
+    },
     err::SydResult,
-    fs::{closeexcept, set_exclusive, set_nonblock},
+    fd::{close, closeexcept, set_exclusive, set_nonblock},
     ignore_signals,
     io::{winsize_get, winsize_set},
     ioctl::IoctlMap,
@@ -41,6 +43,22 @@ use syd::{
     IgnoreSignalOpts,
 };
 
+// Set global allocator to GrapheneOS allocator.
+#[cfg(all(
+    not(coverage),
+    not(feature = "prof"),
+    not(target_os = "android"),
+    target_page_size_4k,
+    target_pointer_width = "64"
+))]
+#[global_allocator]
+static GLOBAL: hardened_malloc::HardenedMalloc = hardened_malloc::HardenedMalloc;
+
+// Set global allocator to tcmalloc if profiling is enabled.
+#[cfg(feature = "prof")]
+#[global_allocator]
+static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
+
 // This is from <linux/tty.h>
 // libc does not export it...
 const N_TTY_BUF_SIZE: usize = 4096;
@@ -81,18 +99,12 @@ syd::main! {
     } = opts;
 
     // SAFETY: Randomize pidfd for hardening.
-    let fpid_fd = duprand(fpid.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-        // SAFETY: duprand returns a valid FD on success.
-        unsafe { OwnedFd::from_raw_fd(fd) }
-    })?;
+    let fpid_fd = duprand(fpid.as_raw_fd(), OFlag::O_CLOEXEC)?;
     drop(fpid);
     let fpid = fpid_fd;
 
     // SAFETY: Randomize PTY fd for hardening.
-    let fpty_fd = duprand(fpty.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-        // SAFETY: duprand returns a valid FD on success.
-        unsafe { OwnedFd::from_raw_fd(fd) }
-    })?;
+    let fpty_fd = duprand(fpty.as_raw_fd(), OFlag::O_CLOEXEC)?;
     drop(fpty);
     let fpty = fpty_fd;
 
@@ -100,10 +112,7 @@ syd::main! {
     let epoll = Epoll::new(EpollCreateFlags::EPOLL_CLOEXEC)?;
 
     // SAFETY: Randomize the epoll fd for hardening.
-    let epoll_fd = duprand(epoll.0.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-        // SAFETY: duprand returns a valid FD on success.
-        unsafe { OwnedFd::from_raw_fd(fd) }
-    })?;
+    let epoll_fd = duprand(epoll.0.as_raw_fd(), OFlag::O_CLOEXEC)?;
     drop(epoll);
     let epoll = Epoll(epoll_fd);
 
@@ -112,38 +121,20 @@ syd::main! {
     // SAFETY: Randomize pipe fds for hardening.
     let (pipe_pty_rd, pipe_pty_wr) = {
         let (rd, wr) = pipe2(OFlag::O_DIRECT | OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
-        let rd = duprand(rd.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
-        })?;
-        let wr = duprand(wr.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
-        })?;
+        let rd = duprand(rd.as_raw_fd(), OFlag::O_CLOEXEC)?;
+        let wr = duprand(wr.as_raw_fd(), OFlag::O_CLOEXEC)?;
         (rd, wr)
     };
     let (pipe_std_rd, pipe_std_wr) = {
         let (rd, wr) = pipe2(OFlag::O_DIRECT | OFlag::O_NONBLOCK | OFlag::O_CLOEXEC)?;
-        let rd = duprand(rd.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
-        })?;
-        let wr = duprand(wr.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
-        })?;
+        let rd = duprand(rd.as_raw_fd(), OFlag::O_CLOEXEC)?;
+        let wr = duprand(wr.as_raw_fd(), OFlag::O_CLOEXEC)?;
         (rd, wr)
     };
 
     // SAFETY: Randomize stdio(3) fds for hardening.
-    let fstd_rd = duprand(libc::STDIN_FILENO, OFlag::O_CLOEXEC).map(|fd| {
-        // SAFETY: duprand returns a valid FD on success.
-        unsafe { OwnedFd::from_raw_fd(fd) }
-    })?;
-    let fstd_wr = duprand(libc::STDOUT_FILENO, OFlag::O_CLOEXEC).map(|fd| {
-        // SAFETY: duprand returns a valid FD on success.
-        unsafe { OwnedFd::from_raw_fd(fd) }
-    })?;
+    let fstd_rd = duprand(libc::STDIN_FILENO, OFlag::O_CLOEXEC)?;
+    let fstd_wr = duprand(libc::STDOUT_FILENO, OFlag::O_CLOEXEC)?;
 
     // SAFETY: Set PTY to exclusive mode to harden against sniffing.
     set_exclusive(&fpty, true)?;
@@ -168,8 +159,8 @@ syd::main! {
     let fsig = {
         let fd = SignalFd::with_flags(&mask, SfdFlags::SFD_NONBLOCK | SfdFlags::SFD_CLOEXEC)?;
         duprand(fd.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { SignalFd::from_owned_fd(OwnedFd::from_raw_fd(fd)) }
+            // SAFETY: dup(3) returned duplicate of a valid signal fd.
+            unsafe { SignalFd::from_owned_fd(fd) }
         })?
     };
 
@@ -179,7 +170,7 @@ syd::main! {
     confine(fsig.as_raw_fd(), debug, print)?;
 
     // Refresh terminal settings.
-    let tio = refresh_pty(&fstd_rd, &fpty);
+    let tio = refresh_pty(&fstd_rd, &fpty)?;
 
     // Refresh window size.
     refresh_win(&fstd_rd, &fpty, ws_x, ws_y);
@@ -204,9 +195,8 @@ syd::main! {
         &pipe_std_wr,
     );
 
-    if let Some(tio) = tio {
-        let _ = tcsetattr(&fstd_rd, SetArg::TCSANOW, &tio);
-    }
+    // Restore terminal settings at exit.
+    tcsetattr(&fstd_rd, SetArg::TCSANOW, &tio)?;
 
     Ok(match result {
         Ok(_) => ExitCode::SUCCESS,
@@ -440,7 +430,7 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
     // They are portable as `struct termios2` has the same size
     // across 32-bit and 64-bit architectures.
     let arch = ScmpArch::native();
-    let ioctl = IoctlMap::new(true);
+    let ioctl = IoctlMap::new(None, true);
     let names = [
         "TCGETS",
         "TCSETS",
@@ -456,7 +446,6 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
             iotty.push(req as u64);
         }
     }
-    drop(ioctl);
 
     #[expect(clippy::disallowed_methods)]
     let syscall = ScmpSyscall::from_name("ioctl").unwrap();
@@ -516,7 +505,7 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
             | CloneFlags::CLONE_NEWNS
             | CloneFlags::CLONE_NEWPID
             | CloneFlags::CLONE_NEWUTS
-            | CloneFlags::from_bits_retain(libc::CLONE_NEWTIME);
+            | CLONE_NEWTIME;
         if unshare(namespaces).is_ok() {
             chroot(".")?; // /proc/self/fdinfo.
             chdir("/")?; // prevent cwd leaking.
@@ -539,6 +528,19 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
 
         // Set the process dumpable attribute to not-dumpable.
         let _ = set_dumpable(false);
+
+        // Set nfiles, nprocs, and filesize rlimits to zero.
+        // Set locks, memory lock and msgqueue rlimits to zero.
+        // Set core dump file size to zero.
+        confine_rlimit_zero(&[
+            Resource::RLIMIT_CORE,
+            Resource::RLIMIT_FSIZE,
+            Resource::RLIMIT_NOFILE,
+            Resource::RLIMIT_NPROC,
+            Resource::RLIMIT_LOCKS,
+            Resource::RLIMIT_MEMLOCK,
+            Resource::RLIMIT_MSGQUEUE,
+        ])?;
     }
 
     if print_rules {
@@ -642,26 +644,22 @@ fn refresh_win<Fd1: AsFd, Fd2: AsFd>(
 }
 
 // Handle terminal settings.
-fn refresh_pty<Fd1: AsFd, Fd2: AsFd>(src: Fd1, dst: Fd2) -> Option<Termios> {
-    let mut tio = if let Ok(tio) = tcgetattr(&src) {
-        tio
-    } else {
-        return None;
-    };
-
-    let orig_tio = tio.clone();
+fn refresh_pty<Fd1: AsFd, Fd2: AsFd>(src: Fd1, dst: Fd2) -> Result<Termios, Errno> {
+    let mut tio = tcgetattr(&src)?;
 
     // Inherit host terminal settings for PTY.
-    let _ = tcsetattr(&dst, SetArg::TCSANOW, &tio);
+    tcsetattr(&dst, SetArg::TCSANOW, &tio)?;
 
     // Set raw mode for input TTY.
+    // Disable background processes from writing.
     // Enable output processing.
+    let orig_tio = tio.clone();
     cfmakeraw(&mut tio);
+    tio.local_flags.insert(LocalFlags::TOSTOP);
     tio.output_flags.insert(OutputFlags::OPOST);
+    tcsetattr(&src, SetArg::TCSANOW, &tio)?;
 
-    let _ = tcsetattr(&src, SetArg::TCSANOW, &tio);
-
-    Some(orig_tio)
+    Ok(orig_tio)
 }
 
 // Parse command line options.
