@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/workers/int.rs: `syd_int' interrupter thread
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -20,7 +20,7 @@ use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
     sched::{unshare, CloneFlags},
-    unistd::{getpid, write, Gid, Pid, Uid},
+    unistd::{getpid, lseek64, write, Gid, Pid, Uid, Whence},
 };
 use serde::{ser::SerializeMap, Serialize, Serializer};
 
@@ -29,12 +29,13 @@ use crate::{
     cache::SysInterrupt,
     config::*,
     confine::{confine_scmp_madvise, confine_scmp_write, scmp_add_setid_rules, ExportMode},
-    cookie::{OPENAT2_COOKIE_ARG4, OPENAT2_COOKIE_ARG5},
+    cookie::{CookieIdx, SYSCOOKIE_POOL},
     err::{err2no, scmp2no, SydJoinHandle, SydResult},
     error,
-    fs::{closeexcept, seccomp_notify_id_valid, tgkill},
+    fd::closeexcept,
+    fs::{seccomp_notify_id_valid, tgkill},
     info,
-    proc::proc_status,
+    proc::{proc_status_open, proc_status_read},
     retry::retry_on_eintr,
     sandbox::Flags,
     sigset::SydSigSet,
@@ -43,6 +44,7 @@ use crate::{
 
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 enum InterruptSource {
+    Manual,
     Signal(SydSigSet),
     ProcessInvalid(Errno),
 }
@@ -54,6 +56,9 @@ impl Serialize for InterruptSource {
     {
         let mut map = serializer.serialize_map(Some(2))?;
         match self {
+            Self::Manual => {
+                map.serialize_entry("name", "manual")?;
+            }
             Self::Signal(set) => {
                 map.serialize_entry("name", "signal")?;
                 map.serialize_entry("set", set)?;
@@ -77,7 +82,7 @@ pub(crate) struct Interrupter {
     transit_gids: Vec<(Gid, Gid)>,
 
     should_exit: Arc<AtomicBool>,
-    cache: Arc<WorkerCache<'static>>,
+    cache: Arc<WorkerCache>,
 }
 
 impl Interrupter {
@@ -87,7 +92,7 @@ impl Interrupter {
         transit_uids: &[(Uid, Uid)],
         transit_gids: &[(Gid, Gid)],
         should_exit: Arc<AtomicBool>,
-        cache: Arc<WorkerCache<'static>>,
+        cache: Arc<WorkerCache>,
     ) -> Self {
         Self {
             flags,
@@ -107,9 +112,9 @@ impl Interrupter {
             .spawn(move || {
                 // SAFETY: We use exit_group(2) here to bail,
                 // because this unsharing is a critical safety feature.
-                if let Err(errno) = unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_FILES) {
+                if let Err(errno) = unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_FILES | CloneFlags::CLONE_SYSVSEM) {
                     alert!("ctx": "boot", "op": "unshare_interrupt_thread",
-                        "msg": format!("failed to unshare(CLONE_FS|CLONE_FILES): {errno}"),
+                        "msg": format!("failed to unshare(CLONE_FS|CLONE_FILES|CLONE_SYSVSEM): {errno}"),
                         "err": errno as i32);
                     std::process::exit(101);
                 }
@@ -175,7 +180,7 @@ impl Interrupter {
                         Ok(ctx) => ctx,
                         Err(error) => {
                             let errno = error.errno().unwrap_or(Errno::ENOSYS);
-                            alert!("ctx": "boot", "op": "confine_interrupt_thread",
+                            alert!("ctx": "boot", "op": "confine_int_thread",
                                 "msg": format!("failed to confine: {error}"),
                                 "err": errno as i32);
                             std::process::exit(101);
@@ -187,7 +192,7 @@ impl Interrupter {
                     // because this confinement is a critical safety feature.
                     if let Err(error) = ctx.load() {
                         let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
-                        alert!("ctx": "boot", "op": "confine_interrupt_thread",
+                        alert!("ctx": "boot", "op": "confine_int_thread",
                             "msg": format!("failed to confine: {error}"),
                             "err": errno as i32);
                         std::process::exit(101);
@@ -196,7 +201,7 @@ impl Interrupter {
                     let safe_setid = self
                         .flags
                         .intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID);
-                    info!("ctx": "confine", "op": "confine_interrupt_thread",
+                    info!("ctx": "confine", "op": "confine_int_thread",
                         "msg": format!("interrupt thread confined with{} SROP mitigation",
                             if safe_setid { "out" } else { "" }));
                 } else {
@@ -213,7 +218,7 @@ impl Interrupter {
     fn main(self, tgid: Pid) -> SydResult<()> {
         loop {
             // Wait for one cycle.
-            std::thread::sleep(INT_CYCLE_TIME);
+            std::thread::sleep(INT_CYCLE_TIME.into());
 
             // Unblock invalidated blocking system calls.
             {
@@ -227,7 +232,10 @@ impl Interrupter {
                     .unwrap_or_else(|err| err.into_inner());
 
                 // Handle interrupts as necessary.
-                map.retain(|_, interrupt| self.handle_interrupt(tgid, interrupt));
+                map.retain_mut(|interrupt| self.handle_interrupt(tgid, interrupt));
+
+                // Keep memory usage minimal.
+                map.shrink_to_fit();
             }
 
             // Check if it's the time to exit.
@@ -239,30 +247,70 @@ impl Interrupter {
         Ok(())
     }
 
-    fn handle_interrupt(&self, tgid: Pid, interrupt: &SysInterrupt) -> bool {
-        // Check pending signals for the thread.
+    // Handles syscall signal interrupts.
+    //
+    // Returns false if interrupt is handled, true otherwise.
+    fn handle_interrupt(&self, tgid: Pid, interrupt: &mut SysInterrupt) -> bool {
+        // Check if syd_emu is already done with the request.
+        if interrupt.delete {
+            return false;
+        }
+        // Check if syd_mon requested manual interruption.
+        if interrupt.signal {
+            Self::interrupt(tgid, interrupt, InterruptSource::Manual);
+            return false;
+        }
+
+        // Open proc_pid_status(5) if it's not open already.
         //
         // SAFETY:
         // 1. We want to wake the respective syd_emu thread in case the
         //    process is no longer valid otherwise we may end up with a
         //    deadlock: See miniupnpc tests, thx kepstin!
         // 2. To prevent PID reuse vectors we validate the request ID.
-        #[expect(clippy::cast_possible_wrap)]
-        let pid = Pid::from_raw(interrupt.request.pid as libc::pid_t);
-        let status = match proc_status(pid) {
-            Ok(status) if self.is_valid(interrupt.request.id) => {
-                // SAFETY: seccomp-id validated, proc_pid_status(5) is valid.
-                status
+        let status_fd = if let Some(fd) = interrupt.status.as_ref() {
+            if let Err(errno) = lseek64(fd, 0, Whence::SeekSet) {
+                if self.is_valid(interrupt.request.id) {
+                    let source = InterruptSource::ProcessInvalid(errno);
+                    Self::interrupt(tgid, interrupt, source);
+                } // no need to interrupt for invalid seccomp-id.
+                return false;
             }
+            fd
+        } else {
+            let fd = match proc_status_open(interrupt.request.pid()) {
+                Ok(fd) if self.is_valid(interrupt.request.id) => {
+                    // seccomp-id validated, proc_pid_status(5) is valid.
+                    fd
+                }
+                Err(errno) if self.is_valid(interrupt.request.id) => {
+                    let source = InterruptSource::ProcessInvalid(errno);
+                    Self::interrupt(tgid, interrupt, source);
+                    return false;
+                }
+                // seccomp-id invalid, no need to interrupt.
+                _ => return false,
+            };
+            interrupt.status = Some(fd);
+            #[expect(clippy::disallowed_methods)]
+            interrupt.status.as_ref().unwrap()
+        };
+
+        // Parse proc_pid_status(5).
+        let status = match proc_status_read(status_fd) {
+            Ok(status) => status,
             Err(errno) if self.is_valid(interrupt.request.id) => {
                 let source = InterruptSource::ProcessInvalid(errno);
                 Self::interrupt(tgid, interrupt, source);
                 return false;
             }
+            // seccomp-id invalid, no need to interrupt.
             _ => return false,
         };
 
-        // Check for per-{thread,process} pending signals:
+        // Check pending signals for the thread.
+        //
+        // 1. Check for per-{thread,process} pending signals:
         //
         // Aggregate pending signals from both the thread-local and
         // process-global queues. In Linux, each thread (LWP) has its
@@ -273,7 +321,7 @@ impl Interrupter {
         // interrupt this execution context.
         let mut sigset = status.sig_pending_thread | status.sig_pending_process;
 
-        // Exclude per-thread blocked signals:
+        // 2. Exclude per-thread blocked signals:
         //
         // Remove those signals currently masked by this thread's signal
         // mask. Signal masks are maintained per-thread via
@@ -288,12 +336,9 @@ impl Interrupter {
             return true;
         }
 
-        // Filter out restarting signals per-process,
-        // unless ignore_restart is set. This may be the
-        // case e.g. when the socket has a timeout for
-        // accept and connect.
-        // Note, `interrupt.ignore_restart` check
-        // was done before calling this function and
+        // 3. Filter out restarting signals per-process, unless ignore_restart is set.
+        // This may be the case e.g. when the socket has a timeout for accept and connect.
+        // Note, `interrupt.ignore_restart` check was done before calling this function and
         // sigset_restart is only Some if it is false.
         if !interrupt.ignore_restart {
             if let Some(sigset_restart) = self
@@ -440,8 +485,8 @@ impl Interrupter {
                     syscall,
                     &[
                         scmp_cmp!($arg0 == PROC_FD() as u64),
-                        scmp_cmp!($arg4 == (*OPENAT2_COOKIE_ARG4).into()),
-                        scmp_cmp!($arg5 == (*OPENAT2_COOKIE_ARG5).into()),
+                        scmp_cmp!($arg4 == SYSCOOKIE_POOL.get(CookieIdx::Openat2Arg4).into()),
+                        scmp_cmp!($arg5 == SYSCOOKIE_POOL.get(CookieIdx::Openat2Arg5).into()),
                     ],
                 )?;
             }
@@ -455,10 +500,18 @@ impl Interrupter {
         // We need this because std::thread::spawn has unwanted
         // side-effects such as opening /sys/devices/system/cpu/online
         // on some architectures.
+        //
+        // Note, we avoid this when profiling is enabled,
+        // as gperf requires it to write profiling data.
         for sysname in ["open", "openat", "stat", "lstat", "statx", "newfstatat"] {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Errno(Errno::ENOSYS as i32), syscall)?;
+                    let action = if !cfg!(feature = "prof") {
+                        ScmpAction::Errno(Errno::ENOSYS as i32)
+                    } else {
+                        ScmpAction::Allow
+                    };
+                    ctx.add_rule(action, syscall)?;
                 }
                 Err(_) => {
                     info!("ctx": "confine", "op": "allow_int_syscall",
@@ -519,34 +572,13 @@ impl Interrupter {
         // Allow safe madvise(2) advice.
         confine_scmp_madvise(&mut ctx)?;
 
-        // Allow safe system calls.
-        for sysname in INT_SYSCALLS.iter().chain(VDSO_SYSCALLS) {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_int_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
-
-        // Allow futex system calls.
-        for sysname in FUTEX_SYSCALLS {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_int_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
-
-        // Allow getid system calls.
-        for sysname in GET_ID_SYSCALLS {
+        // Allow safe, futex and getid system calls.
+        for sysname in INT_SYSCALLS
+            .iter()
+            .chain(FUTEX_SYSCALLS)
+            .chain(GET_ID_SYSCALLS)
+            .chain(VDSO_SYSCALLS)
+        {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
                     ctx.add_rule(ScmpAction::Allow, syscall)?;
