@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/workers/ipc.rs: `syd_ipc' remote-configuration thread
 //
-// Copyright (c) 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -30,7 +30,7 @@ use nix::{
         epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
         socket::{
             accept4, bind, listen, recvmsg, sendmsg, setsockopt, socket, sockopt::PassCred,
-            AddressFamily, Backlog, ControlMessageOwned, MsgFlags, SockFlag, SockType, UnixAddr,
+            AddressFamily, Backlog, ControlMessageOwned, SockFlag, SockType, UnixAddr,
             UnixCredentials,
         },
         stat::{umask, Mode},
@@ -41,15 +41,18 @@ use serde::{ser::SerializeMap, Serialize, Serializer};
 
 use crate::{
     alert,
-    compat::{epoll_ctl_mod_safe, epoll_ctl_safe},
+    compat::{epoll_ctl_mod_safe, epoll_ctl_safe, MsgFlags},
     config::*,
-    confine::{confine_scmp_write, scmp_add_setid_rules},
+    confine::{confine_scmp_madvise, confine_scmp_write, scmp_add_setid_rules},
     err::{err2no, scmp2no, SydJoinHandle, SydResult},
-    fs::{closeexcept, peer_cred, retry_on_eintr},
+    fd::closeexcept,
+    fs::peer_cred,
     hash::SydHashMap,
     info,
+    landlock::Errata,
     landlock_policy::LandlockPolicy,
     path::{XPath, XPathBuf},
+    retry::retry_on_eintr,
     rng::duprand,
     sandbox::{Flags, Sandbox, LINE_MAX},
 };
@@ -269,14 +272,11 @@ impl IpcWorker {
         )?;
 
         // SAFETY: Randomize to make fd reuse harder.
-        let sock_fd = duprand(sock.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
+        let sock_fd = duprand(sock.as_raw_fd(), OFlag::O_CLOEXEC).inspect(|fd| {
             // SAFETY: Used to prevent leaks in sandbox process:
             let mut fd_str = itoa::Buffer::new();
-            let fd_str = fd_str.format(fd);
+            let fd_str = fd_str.format(fd.as_raw_fd());
             std::env::set_var(ENV_IPC_UNIX_FD, fd_str);
-
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
         })?;
         drop(sock);
         let sock = sock_fd;
@@ -296,14 +296,11 @@ impl IpcWorker {
         let epoll = Epoll::new(EpollCreateFlags::EPOLL_CLOEXEC)?;
 
         // SAFETY: Randomize the epoll fd to make fd reuse harder.
-        let epoll_fd = duprand(epoll.0.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
+        let epoll_fd = duprand(epoll.0.as_raw_fd(), OFlag::O_CLOEXEC).inspect(|fd| {
             // SAFETY: Used to prevent leaks in sandbox process:
             let mut fd_str = itoa::Buffer::new();
-            let fd_str = fd_str.format(fd);
+            let fd_str = fd_str.format(fd.as_raw_fd());
             std::env::set_var(ENV_IPC_POLL_FD, fd_str);
-
-            // SAFETY: duprand returns a valid FD on success.
-            unsafe { OwnedFd::from_raw_fd(fd) }
         })?;
         drop(epoll);
         let epoll = Epoll(epoll_fd);
@@ -338,12 +335,11 @@ impl IpcWorker {
     ) -> SydResult<ScmpFilterContext> {
         if !dry_run {
             // SAFETY: Set up a Landlock sandbox to disallow all access.
-            // We do not enable scoped signals not to break inter-thread
-            // signaling. This was fixed in recent versions of Landlock.
-            // TODO: Make LandlockPolicy use scoped signals only if fixed.
             let abi = crate::landlock::ABI::new_current();
+            let errata = crate::landlock::Errata::query();
             let policy = LandlockPolicy {
                 scoped_abs: true,
+                scoped_sig: errata.contains(Errata::SCOPED_SIGNAL_SAME_TGID),
                 ..Default::default()
             };
             let _ = policy.restrict_self(abi);
@@ -361,8 +357,8 @@ impl IpcWorker {
         ctx.set_ctl_tsync(false)?;
 
         // Disable Speculative Store Bypass mitigations
-        // with trace/allow_unsafe_spec_exec:1
-        ctx.set_ctl_ssb(flags.allow_unsafe_spec_exec())?;
+        // with trace/allow_unsafe_exec_speculative:1
+        ctx.set_ctl_ssb(flags.allow_unsafe_exec_speculative())?;
 
         // We kill for bad system call and bad arch.
         ctx.set_act_badarch(ScmpAction::KillProcess)?;
@@ -396,6 +392,9 @@ impl IpcWorker {
         // Deny open and stat family with ENOSYS rather than KillProcess.
         // We need this because Rust allocator has side effects such as
         // opening /proc/sys/vm/overcommit_memory on some architectures.
+        //
+        // Note, we avoid this when profiling is enabled,
+        // as gperf requires it to write profiling data.
         for sysname in [
             "open",
             "openat",
@@ -407,7 +406,12 @@ impl IpcWorker {
         ] {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Errno(Errno::ENOSYS as i32), syscall)?;
+                    let action = if !cfg!(feature = "prof") {
+                        ScmpAction::Errno(Errno::ENOSYS as i32)
+                    } else {
+                        ScmpAction::Allow
+                    };
+                    ctx.add_rule(action, syscall)?;
                 }
                 Err(_) => {
                     info!("ctx": "confine", "op": "allow_ipc_syscall",
@@ -465,6 +469,9 @@ impl IpcWorker {
         // No proc_pid_mem(5) access required here.
         confine_scmp_write(&mut ctx, None, false)?;
 
+        // Allow safe madvise(2) advice.
+        confine_scmp_madvise(&mut ctx)?;
+
         // Allow safe system calls.
         for sysname in IPC_SYSCALLS.iter().chain(VDSO_SYSCALLS) {
             match ScmpSyscall::from_name(sysname) {
@@ -537,9 +544,11 @@ impl IpcWorker {
     fn init(&mut self, notif_pipe: (RawFd, RawFd)) -> SydResult<()> {
         // SAFETY: We use exit_group(2) here to bail,
         // because this unsharing is a critical safety feature.
-        if let Err(errno) = unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_FILES) {
+        if let Err(errno) =
+            unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_FILES | CloneFlags::CLONE_SYSVSEM)
+        {
             alert!("ctx": "boot", "op": "unshare_ipc_thread",
-                "msg": format!("failed to unshare(CLONE_FS|CLONE_FILES): {errno}"),
+                "msg": format!("failed to unshare(CLONE_FS|CLONE_FILES|CLONE_SYSVSEM): {errno}"),
                 "err": errno as i32);
             std::process::exit(101);
         }
@@ -733,7 +742,7 @@ impl IpcWorker {
                     conn.stream.as_raw_fd(),
                     &mut iov,
                     Some(&mut cmsg),
-                    MsgFlags::MSG_DONTWAIT,
+                    MsgFlags::MSG_DONTWAIT.into(),
                 ) {
                     Ok(msg) if msg.bytes == 0 => {
                         // EOF on read side: if there's a pending write,
@@ -868,7 +877,7 @@ impl IpcWorker {
                     conn.stream.as_raw_fd(),
                     &iov,
                     &[],
-                    MsgFlags::MSG_DONTWAIT,
+                    MsgFlags::MSG_DONTWAIT.into(),
                     None,
                 ) {
                     Ok(0) => {
